@@ -102,7 +102,7 @@ impl Debug for Part {
 /// For each node, it contains `proposal_index, receiver_index, serialised value for the receiver,
 /// encrypted values from the sender`.
 #[derive(Deserialize, Serialize, Clone, Hash, Eq, PartialEq, PartialOrd, Ord)]
-pub struct Acknowledgment(u64, u64, Vec<u8>, Vec<Vec<u8>>);
+pub struct Acknowledgment(pub u64, u64, Vec<u8>, Vec<Vec<u8>>);
 
 impl Debug for Acknowledgment {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
@@ -281,6 +281,8 @@ impl ComplaintsAccumulator {
     }
 }
 
+pub type MessageAndTarget = (XorName, Message);
+
 /// An algorithm for dealerless distributed key generation.
 ///
 /// This is trying to follow the protocol as suggested at
@@ -318,8 +320,8 @@ pub struct KeyGen {
     complaints_accumulator: ComplaintsAccumulator,
     /// Pending complain messages.
     pending_complain_messages: Vec<Message>,
-    /// Pending messages that cannot handle yet.
-    pending_messages: Vec<Message>,
+    /// Cached messages to be used for reply unhandable.
+    message_cache: BTreeMap<XorName, Message>,
 }
 
 impl KeyGen {
@@ -329,7 +331,7 @@ impl KeyGen {
         our_id: XorName,
         threshold: usize,
         names: BTreeSet<XorName>,
-    ) -> Result<(KeyGen, Message), Error> {
+    ) -> Result<(KeyGen, Vec<MessageAndTarget>), Error> {
         if names.len() < threshold {
             return Err(Error::Unknown);
         }
@@ -350,18 +352,18 @@ impl KeyGen {
             initalization_accumulator: InitializationAccumulator::new(),
             complaints_accumulator: ComplaintsAccumulator::new(names.clone(), threshold),
             pending_complain_messages: Vec::new(),
-            pending_messages: Vec::new(),
+            message_cache: BTreeMap::new(),
         };
 
-        Ok((
-            key_gen,
-            Message::Initialization {
-                key_gen_id: our_index,
-                m: threshold,
-                n: names.len(),
-                member_list: names,
-            },
-        ))
+        let msg = Message::Initialization {
+            key_gen_id: our_index,
+            m: threshold,
+            n: names.len(),
+            member_list: names.clone(),
+        };
+        let messages: Vec<_> = names.iter().map(|name| (*name, msg.clone())).collect();
+
+        Ok((key_gen, messages))
     }
 
     pub fn phase(&self) -> Phase {
@@ -373,56 +375,112 @@ impl KeyGen {
         &mut self,
         rng: &mut R,
         msg: Message,
-    ) -> Result<Vec<Message>, Error> {
+    ) -> Result<Vec<MessageAndTarget>, Error> {
         if self.is_finalized() {
             return Ok(Vec::new());
         }
-        let result = self.process_message(rng, msg.clone());
-        match result {
-            Ok(mut msgs) => {
-                msgs.extend(self.poll_pending_messages(rng));
-                Ok(msgs)
-            }
-            Err(Error::UnexpectedPhase { .. }) | Err(Error::MissingPart) => {
-                self.pending_messages.push(msg);
-                Ok(Vec::new())
-            }
-            Err(_) => result,
-        }
+
+        self.process_message(rng, msg)
     }
 
-    fn poll_pending_messages<R: RngCore>(&mut self, rng: &mut R) -> Vec<Message> {
-        let pending_messages = std::mem::take(&mut self.pending_messages);
+    /// Cached message will be returned as a list,
+    /// with Initialization messages on top and Proposal behind.
+    /// They shall get handled in such order on receiver side as well.
+    pub fn get_cached_message(&self) -> Vec<Message> {
+        let mut result = Vec::new();
+        result.extend(
+            self.message_cache
+                .iter()
+                .filter_map(|(_, msg)| match msg {
+                    Message::Initialization { .. } => Some(msg.clone()),
+                    _ => None,
+                })
+                .collect::<Vec<Message>>(),
+        );
+        result.extend(
+            self.message_cache
+                .iter()
+                .filter_map(|(_, msg)| match msg {
+                    Message::Proposal { .. } => Some(msg.clone()),
+                    _ => None,
+                })
+                .collect::<Vec<Message>>(),
+        );
+        result
+    }
+
+    /// Handle upper layer cached messages even before this DKG session got started.
+    /// Returns with messages need to be broadcast to all,
+    /// AND unhandable messages need to be sent to the creator.
+    /// It is also being used when handle message_history due to unhandable.
+    pub fn handle_pre_session_messages<R: RngCore>(
+        &mut self,
+        rng: &mut R,
+        mut cache_messages: Vec<Message>,
+    ) -> (Vec<MessageAndTarget>, Vec<MessageAndTarget>) {
         let mut msgs = Vec::new();
-        for message in pending_messages {
-            if let Ok(new_messages) = self.process_message(rng, message.clone()) {
-                if self.is_finalized() {
-                    return Vec::new();
+        let mut updated = false;
+        loop {
+            trace!("new round polling history messages");
+            let pending_messages = std::mem::take(&mut cache_messages);
+            for message in pending_messages {
+                if let Ok(new_messages) = self.process_message(rng, message.clone()) {
+                    if self.is_finalized() {
+                        return (Vec::new(), Vec::new());
+                    }
+                    msgs.extend(new_messages);
+                    updated = true;
+                } else {
+                    trace!("pushing back history message {:?}", message);
+                    cache_messages.push(message);
                 }
-                msgs.extend(new_messages);
+            }
+            if !updated {
+                break;
             } else {
-                self.pending_messages.push(message);
+                updated = false;
             }
         }
-        msgs
+
+        let mut unhandables = Vec::new();
+        for msg in cache_messages {
+            let sender = if let Some(name) = self.node_id_from_index(msg.creator()) {
+                name
+            } else {
+                warn!(
+                    "cannot get name of index {:?} among {:?}",
+                    msg.creator(),
+                    self.names
+                );
+                continue;
+            };
+            unhandables.push((sender, msg));
+        }
+        (msgs, unhandables)
     }
 
     fn process_message<R: RngCore>(
         &mut self,
         rng: &mut R,
         msg: Message,
-    ) -> Result<Vec<Message>, Error> {
-        debug!(
-            "{:?} with phase {:?} handle DKG message {:?}",
-            self, self.phase, msg
+    ) -> Result<Vec<MessageAndTarget>, Error> {
+        trace!(
+            "{:?} with phase {:?} handle DKG message {:?}-{:?}",
+            self,
+            self.phase,
+            msg.id(),
+            msg
         );
-        match msg {
+        let result = match msg.clone() {
             Message::Initialization {
                 key_gen_id,
                 m,
                 n,
                 member_list,
-            } => self.handle_initialization(rng, m, n, key_gen_id, member_list),
+            } => {
+                let _ = self.message_cache.insert(msg.id(), msg);
+                self.handle_initialization(rng, m, n, key_gen_id, member_list)
+            }
             Message::Proposal { key_gen_id, part } => self.handle_proposal(key_gen_id, part),
             Message::Complaint {
                 key_gen_id,
@@ -434,7 +492,8 @@ impl KeyGen {
                 keys_map,
             } => self.handle_justification(key_gen_id, keys_map),
             Message::Acknowledgment { key_gen_id, ack } => self.handle_ack(key_gen_id, ack),
-        }
+        };
+        self.multicasting_messages(result?)
     }
 
     // Handles an incoming initialize message. Creates the `Proposal` message once quorumn
@@ -448,10 +507,7 @@ impl KeyGen {
         member_list: BTreeSet<XorName>,
     ) -> Result<Vec<Message>, Error> {
         if self.phase != Phase::Initialization {
-            return Err(Error::UnexpectedPhase {
-                expected: Phase::Initialization,
-                actual: self.phase,
-            });
+            return Ok(Vec::new());
         }
 
         if let Some((m, _n, member_list)) =
@@ -501,11 +557,13 @@ impl KeyGen {
     // When there is an invalidation happens, holds the `Complaint` message till broadcast out
     // when `finalize_contributing` being called.
     fn handle_proposal(&mut self, sender_index: u64, part: Part) -> Result<Vec<Message>, Error> {
-        if !(self.phase == Phase::Contribution || self.phase == Phase::Commitment) {
+        if self.phase == Phase::Initialization {
             return Err(Error::UnexpectedPhase {
                 expected: Phase::Contribution,
                 actual: self.phase,
             });
+        } else if !(self.phase == Phase::Contribution || self.phase == Phase::Commitment) {
+            return Ok(Vec::new());
         }
 
         let row = match self.handle_part_or_fault(sender_index, part.clone()) {
@@ -517,7 +575,7 @@ impl KeyGen {
                     part,
                 };
                 debug!(
-                    "{:?} complain {:?} with Error {:?}",
+                    "{:?} complain {:?} with Error {:?} when handling a proposal",
                     self, sender_index, _fault
                 );
                 let invalid_contribute = serialize(&msg)?;
@@ -565,19 +623,22 @@ impl KeyGen {
         sender_index: u64,
         ack: Acknowledgment,
     ) -> Result<Vec<Message>, Error> {
-        if !(self.phase == Phase::Contribution || self.phase == Phase::Commitment) {
+        if self.phase == Phase::Initialization {
             return Err(Error::UnexpectedPhase {
                 expected: Phase::Contribution,
                 actual: self.phase,
             });
+        } else if !(self.phase == Phase::Contribution || self.phase == Phase::Commitment) {
+            return Ok(Vec::new());
         }
+
         match self.handle_ack_or_fault(sender_index, ack.clone()) {
             Ok(()) => {
                 if self.all_contribution_received() {
                     if self.phase == Phase::Commitment {
                         self.become_finalization();
                     } else {
-                        return self.finalize_contributing_phase();
+                        return Ok(self.finalize_contributing_phase());
                     }
                 }
             }
@@ -618,7 +679,7 @@ impl KeyGen {
                 .all(|part| part.acks.len() == self.names.len())
     }
 
-    fn finalize_contributing_phase(&mut self) -> Result<Vec<Message>, Error> {
+    fn finalize_contributing_phase(&mut self) -> Vec<Message> {
         self.phase = Phase::Complaining;
 
         for non_contributor in self.non_contributors().0 {
@@ -645,8 +706,7 @@ impl KeyGen {
             self.become_finalization();
         }
 
-        self.pending_messages.clear();
-        Ok(mem::take(&mut self.pending_complain_messages))
+        mem::take(&mut self.pending_complain_messages)
     }
 
     fn non_contributors(&self) -> (BTreeSet<u64>, BTreeSet<XorName>) {
@@ -678,23 +738,11 @@ impl KeyGen {
     pub fn timed_phase_transition<R: RngCore>(
         &mut self,
         rng: &mut R,
-    ) -> Result<Vec<Message>, Error> {
-        debug!("{:?} current phase is {:?}", self, self.phase);
-        match self.phase {
-            Phase::Contribution => match self.finalize_contributing_phase() {
-                Ok(mut messages) => {
-                    messages.extend(self.poll_pending_messages(rng));
-                    Ok(messages)
-                }
-                Err(err) => Err(err),
-            },
-            Phase::Complaining => match self.finalize_complaining_phase(rng) {
-                Ok(mut messages) => {
-                    messages.extend(self.poll_pending_messages(rng));
-                    Ok(messages)
-                }
-                Err(err) => Err(err),
-            },
+    ) -> Result<Vec<MessageAndTarget>, Error> {
+        trace!("{:?} current phase is {:?}", self, self.phase);
+        let result = match self.phase {
+            Phase::Contribution => Ok(self.finalize_contributing_phase()),
+            Phase::Complaining => self.finalize_complaining_phase(rng),
             Phase::Initialization => Err(Error::UnexpectedPhase {
                 expected: Phase::Contribution,
                 actual: self.phase,
@@ -705,7 +753,55 @@ impl KeyGen {
             }),
 
             Phase::Finalization => Ok(Vec::new()),
+        };
+        self.multicasting_messages(result?)
+    }
+
+    // Specify the receiver of the DKG messages explicitly
+    // to avoid un-necessary broadcasting.
+    fn multicasting_messages(
+        &mut self,
+        messages: Vec<Message>,
+    ) -> Result<Vec<MessageAndTarget>, Error> {
+        let mut messaging = Vec::new();
+        for message in messages {
+            match message {
+                Message::Proposal { ref part, .. } => {
+                    // Proposal to us cannot be used by other.
+                    // So the cache must be carried out on sender side.
+                    let _ = self.message_cache.insert(message.id(), message.clone());
+
+                    let receiver = if let Some(name) = self.node_id_from_index(part.receiver) {
+                        name
+                    } else {
+                        warn!(
+                            "For a Proposal, Cannot get name of index {:?} among {:?}",
+                            part.receiver, self.names
+                        );
+                        continue;
+                    };
+                    messaging.push((receiver, message));
+                }
+                Message::Acknowledgment { ref ack, .. } => {
+                    let receiver = if let Some(name) = self.node_id_from_index(ack.1) {
+                        name
+                    } else {
+                        warn!(
+                            "For an Acknowledgement, Cannot get name of index {:?} among {:?}",
+                            ack.1, self.names
+                        );
+                        continue;
+                    };
+                    messaging.push((receiver, message));
+                }
+                _ => {
+                    for name in &self.names {
+                        messaging.push((*name, message.clone()));
+                    }
+                }
+            }
         }
+        Ok(messaging)
     }
 
     // Handles a `Complaint` message.
@@ -716,10 +812,8 @@ impl KeyGen {
         invalid_msg: Vec<u8>,
     ) -> Result<Vec<Message>, Error> {
         if self.phase != Phase::Complaining {
-            return Err(Error::UnexpectedPhase {
-                expected: Phase::Complaining,
-                actual: self.phase,
-            });
+            trace!("To avoid triggering AE pattern, skip this so far");
+            return Ok(Vec::new());
         }
 
         let sender_id = self
@@ -739,6 +833,7 @@ impl KeyGen {
         rng: &mut R,
     ) -> Result<Vec<Message>, Error> {
         let failings = self.complaints_accumulator.finalize_complaining_phase();
+
         if failings.len() >= self.names.len() - self.threshold {
             let mut result = BTreeSet::new();
             failings.iter().for_each(|pk| {
@@ -746,6 +841,7 @@ impl KeyGen {
                     let _ = result.insert(index);
                 }
             });
+            trace!("Finalized with too many failing voters");
             return Err(Error::TooManyNonVoters(result));
         }
 
@@ -821,7 +917,6 @@ impl KeyGen {
 
     fn become_finalization(&mut self) {
         self.phase = Phase::Finalization;
-        self.pending_messages.clear();
         self.pending_complain_messages.clear();
     }
 
@@ -834,7 +929,7 @@ impl KeyGen {
     }
 
     /// Returns the id of the index, or `None` if it is unknown.
-    fn node_id_from_index(&self, node_index: u64) -> Option<XorName> {
+    pub fn node_id_from_index(&self, node_index: u64) -> Option<XorName> {
         for (i, name) in self.names.iter().enumerate() {
             if i == node_index as usize {
                 return Some(*name);
@@ -859,7 +954,16 @@ impl KeyGen {
 
     /// Returns `true` if in the phase of Finalization.
     pub fn is_finalized(&self) -> bool {
-        self.phase == Phase::Finalization
+        let result = self.phase == Phase::Finalization;
+
+        if !result {
+            trace!("incompleted DKG session containing:");
+            for (key, part) in self.parts.iter() {
+                let acks: Vec<u64> = part.values.keys().cloned().collect();
+                trace!("    Part from {:?}, and acks from {:?}", key, acks);
+            }
+        }
+        result
     }
 
     /// Returns the new secret key share and the public key set.
@@ -1037,7 +1141,7 @@ impl KeyGen {
             initalization_accumulator: InitializationAccumulator::new(),
             complaints_accumulator: ComplaintsAccumulator::new(names, threshold),
             pending_complain_messages: Vec::new(),
-            pending_messages: Vec::new(),
+            message_cache: BTreeMap::new(),
         }
     }
 }
